#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['manifoldy', 'manifoldy.test']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.18', 'scipy>=1.4', 'ndsplines>=0.1.2', 'joblib>=1.0.1']

extras_require = \
{'doc': ['sphinx'],
 'test': ['pytest >=2.7.3',
          'pytest-cov',
          'numpy>=1.18',
          'scipy>=1.4',
          'ndsplines>=0.1.2',
          'joblib>=1.0.1']}

setup(name='manifoldy',
      version='0.5.0',
      description='A package to generate customizable manifold synthetic datasets.',
      author='Raul Lara-Cabrera',
      author_email='raul.lara@upm.es',
      url='https://github.com/KNODIS-Research-Group/manifoldy',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.8',
     )
